@@ -72,6 +72,15 @@
 #ifndef NO_NAMESPACE_GIAC
 namespace giac {
 #endif // ndef NO_NAMESPACE_GIAC
+  void wait_1ms(int ms=1);
+  template<class U>
+  inline unsigned sizeinbase2(U n){
+    unsigned i=0;
+    for (;n;++i){
+      n >>= 1;
+    }
+    return i;
+  }
   std::vector<int> operator % (const std::vector<int> & a,int modulo);
   std::vector<int> operator / (const std::vector<int> & v,const std::vector<int> & b);
   std::vector<int> operator % (const std::vector<int> & v,const std::vector<int> & b);
@@ -235,6 +244,31 @@ namespace giac {
     }
   }
 
+#ifdef INT128
+  // tmp is an allocated mpz_t 
+  inline void mpz2int128(mpz_t * ptr,mpz_t * tmp,int128_t & ans){
+    int i=mpz_sgn(*ptr);
+    if (i<0)
+      mpz_neg(*ptr,*ptr);
+    mpz_tdiv_q_2exp(*tmp,*ptr,93);
+    ans=mpz_get_ui(*tmp);
+    ans <<= 93;
+    mpz_tdiv_q_2exp(*tmp,*ptr,62);
+    mpz_tdiv_r_2exp(*tmp,*tmp,31);
+    ans += (int128_t(mpz_get_ui(*tmp)) << 62);
+    mpz_tdiv_q_2exp(*tmp,*ptr,31);
+    mpz_tdiv_r_2exp(*tmp,*tmp,31);
+    ans += (ulonglong(mpz_get_ui(*tmp))<<31);
+    mpz_tdiv_r_2exp(*tmp,*ptr,31);
+    ans += mpz_get_ui(*tmp);
+    if (i<0){
+      mpz_neg(*ptr,*ptr);
+      ans = -ans;
+    }
+  }
+
+#endif
+
   class my_mpz {
   public:
     mpz_t ptr;
@@ -454,6 +488,116 @@ namespace giac {
     unsigned int _capacity;
   };
 
+#if 0
+  template<class T>
+  class vector_size32 {
+  public:
+    struct {
+      unsigned int taille;
+      T v0;
+      T v1;
+      T v2;
+      T v3;
+      T v4;
+      T v5;
+      T v6;
+    };
+    vector_size32(){ taille=1;}
+    vector_size32(size_t t){ 
+      if (t>7){
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	mytabptr->tab = new T[t];
+	mytabptr->_size = mytabptr->_capacity = t;
+	int * ptr=(int *)mytabptr->tab;
+	for (size_t i=0;i<t;++ptr,++i)
+	  *ptr=0;
+      }
+      else { 
+	taille=2*t+1; 
+	* (int *) &v0=0; * (longlong *) &v1=0;
+	* (longlong *) &v3=0;
+	* (longlong *) &v5=0;
+      }
+    }
+    vector_size32(size_t t,const T & elem){ 
+      if (t>7){ 
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	mytabptr->tab = new T[t];
+	mytabptr->_size = mytabptr->_capacity = t;
+	T * ptr=mytabptr->tab;
+	for (size_t i=0;i<t;++ptr,++i)
+	  *ptr=elem;
+      }
+      else { 
+	taille=2*t+1; 
+	v6=v5=v4=v3=v2=v1=v0=elem;
+      }
+    }
+    ~vector_size32(){
+      if (taille%2==0){
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	delete [] mytabptr->tab;
+      }
+    }
+    void push_back(T t){
+      if (taille %2){
+	if (taille==2*7+1){
+	  T * ptr = new T[16];
+	  *ptr=v0; ++ptr;
+	  *ptr=v1; ++ptr;
+	  *ptr=v2; ++ptr;
+	  *ptr=v3; ++ptr;
+	  *ptr=v4; ++ptr;
+	  *ptr=v5; ++ptr;
+	  *ptr=v6; ++ptr;
+	  *ptr=t;
+	  mytab<T> * mytabptr = (mytab<T> *) this;
+	  mytabptr->tab = ptr-7;
+	  mytabptr->_size = 8;
+	  mytabptr->_capacity = 16;
+	  return;
+	}
+	*(&v0+taille/2)=t;
+	taille += 2;
+      }
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	if (mytabptr->_size>=mytabptr->_capacity){
+	  mytabptr->_capacity *=2;
+	  T * ptr = new T[mytabptr->_capacity];
+	  memcpy(ptr,mytabptr->tab,mytabptr->_size*sizeof(T));
+	  delete [] mytabptr->tab;
+	  mytabptr->tab=ptr;
+	}
+	mytabptr->tab[mytabptr->_size]=t;
+	++mytabptr->_size;
+      }
+    }
+    void pop_back(){
+      if (taille%2)
+	taille -=2;
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	--mytabptr->_size;
+      }
+    }
+    void clear(){
+      if (taille%2)
+	taille=1;
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	mytabptr->_size=0;
+      }
+    }
+    T operator [] (size_t pos) const {
+      if (taille%2)
+	return *(&v0+pos);
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	return mytabptr->tab[pos];
+      }
+    }
+#else
   template<class T>
   class vector_size32 {
   public:
@@ -506,12 +650,8 @@ namespace giac {
 	  mytabptr->_capacity = 6;
 	  return;
 	}
+	*(&v0+taille/2)=t;
 	taille += 2;
-	if (taille==7)
-	  v2=t;
-	else { 
-	  if (taille==5) v1=t; else v0=t; 
-	}
       }
       else {
 	mytab<T> * mytabptr = (mytab<T> *) this;
@@ -544,12 +684,13 @@ namespace giac {
     }
     T operator [] (size_t pos) const {
       if (taille%2)
-	return pos?((pos==2)?v2:v1):v0;
+	return *(&v0+pos); // pos?((pos==2)?v2:v1):v0;
       else {
 	mytab<T> * mytabptr = (mytab<T> *) this;
 	return mytabptr->tab[pos];
       }
     }
+#endif
     typename std::vector<T>::const_iterator begin() const {
       if (taille %2)
 	return typename std::vector<T>::const_iterator(&v0);
@@ -640,8 +781,8 @@ namespace giac {
       v.push_back(*it2);
   }
 
-  template<class T,class U>
-  void smalladd(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,int reduce){
+  template<class T,class U,class R>
+  void smalladd(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,const R & reduce){
     if (&v1==&v || &v2==&v){
       std::vector< T_unsigned<T,U> > tmp;
       smalladd(v1,v2,tmp,reduce);
@@ -842,8 +983,8 @@ namespace giac {
   // (powmod should be defined for T,U,T)
   // * should not overflow in T, if T=int, reduce must be < 46340
   // this could be fixed using type_operator_... instead of *
-  template<class T,class U>
-  void smallhorner(const std::vector< T_unsigned<T,U> > & v1,const T & g,const std::vector<U> & vars,std::vector< T_unsigned<T,U> > & v,int reduce){
+  template<class T,class U,class R>
+  void smallhorner(const std::vector< T_unsigned<T,U> > & v1,const T & g,const std::vector<U> & vars,std::vector< T_unsigned<T,U> > & v,const R& reduce){
     typename std::vector< T_unsigned<T,U> >::const_iterator it=v1.begin(),itend=v1.end();
     U deg=vars.back();
     v.clear();
@@ -930,8 +1071,8 @@ namespace giac {
   template<class T,class U,class R>
   void smallmult(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,const R & reduce,size_t possible_size);
 
-  template<class T,class U>
-  void smallmulpoly_interpolate(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,const std::vector<U> & vars,const index_t & vdeg,int reduce){
+  template<class T,class U,class R>
+  void smallmulpoly_interpolate(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,const std::vector<U> & vars,const index_t & vdeg,const R& reduce){
     int dim=int(vars.size());
     if (dim==1){
       smallmult(v1,v2,v,reduce,0);
@@ -960,10 +1101,10 @@ namespace giac {
     }
     // interpolation
     for (int alpha=s-1;alpha>=0;--alpha){
-      smallmult<T,U>(-alpha,v,tmp2,reduce); 
+      smallmult<T,U,R>(-alpha,v,tmp2,reduce); 
       smallshift(v,U(1),v); // multiply v*(x-alpha)
-      smalladd<T,U>(v,tmp2,tmp3,reduce);
-      smalladd<T,U>(tmp3,tab[alpha],v,reduce);
+      smalladd<T,U,R>(v,tmp2,tmp3,reduce);
+      smalladd<T,U,R>(tmp3,tab[alpha],v,reduce);
     }
     delete [] tab;
   }
@@ -1004,8 +1145,8 @@ namespace giac {
     delete [] tab;
   }
 
-  template<class T,class U>
-  void smallmulpoly_interpolate(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,const index_t & vdeg,int reduce){
+  template<class T,class U,class R>
+  void smallmulpoly_interpolate(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,const index_t & vdeg,const R & reduce){
     int dim=int(vdeg.size());
     std::vector<U> vars(dim);
     vars.back()=vdeg.back();
@@ -1044,7 +1185,7 @@ namespace giac {
   }
 
   template<class T>
-  void in_out_heap(T * tab,unsigned size,T value){
+  void in_out_heap(T * tab,size_t size,T value){
     unsigned childindex=2,holeindex=0;
     while (childindex<size){
       // find largest child until end of tab
@@ -1185,7 +1326,9 @@ namespace giac {
 	// using heap of chains
 	// std::vector< vector_size64< std::pair<unsigned,unsigned> > > vindex(v1s);
 	std::vector< std::vector< std::pair<unsigned,unsigned> > > vindex(v1s);
+#ifdef HEAP_STATS
 	double count1=0,count2=0,total=double(v1s)*v2s;
+#endif
 	U_unsigned<U> * heap = new U_unsigned<U>[v1s] ; // pointers to v2 monomials
 	U_unsigned<U> * heap0, *heapbeg=heap,* heapend=heap+v1s;
 	for (it1=it1beg,heap0=heap;heap0!=heapend;++heap0,++it1){
@@ -1210,10 +1353,11 @@ namespace giac {
 	    it=vindex[heapbeg->v].begin();
 	    itend=vindex[heapbeg->v].end();
 	    for (;it!=itend;++it){
-	      type_operator_plus_times_reduce((it1beg+it->first)->g,(it2beg+it->second)->g,g,reduce);
+	      unsigned & its=it->second;
+	      type_operator_plus_times_reduce((it1beg+it->first)->g,(it2beg+its)->g,g,reduce);
 	      // increment 2nd poly index of the elements of the top chain
-	      ++it->second;
-	      if (it->second!=v2s)
+	      ++its;
+	      if (its!=v2s)
 		nouveau.push_back(*it);
 	    }
 #ifdef USTL
@@ -1228,32 +1372,46 @@ namespace giac {
 	  // erase top node, then push each element of the incremented top chain 
 	  it=nouveau.begin();
 	  itend=nouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    u=(it1beg+it->first)->u+(it2beg+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vindex[previndex].push_back(*it);
+#ifdef HEAP_STATS
+	      ++count1;
+#endif
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vindex[heapbeg->v].push_back(*it);
+	      vindex[previndex=heapbeg->v].push_back(*it);
+#ifdef HEAP_STATS
 	      ++count1;
+#endif
 	      continue;
 	    }
 	    bool done=false;
 	    while (holeindex){
 	      parentindex=(holeindex-1) >> 1;
-	      if (u<(heapbeg+parentindex)->u)
+	      U pu=(heapbeg+parentindex)->u;
+	      if (u<pu)
 		break;
-	      if (u==(heapbeg+parentindex)->u){
+	      if (u==pu){
 		done=true;
+#ifdef HEAP_STATS
 		++count2;
-		vindex[(heapbeg+parentindex)->v].push_back(*it);
+#endif
+		vindex[previndex=(heapbeg+parentindex)->v].push_back(*it);
 		break;
 	      }
 	      holeindex=parentindex;
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vindex[heapend->v].clear();
-	      vindex[heapend->v].push_back(*it);
+	      vindex[previndex=heapend->v].clear();
+	      vindex[previndex].push_back(*it);
 	      heapend->u=u;
 	      ++heapend;
 #ifdef USTL
@@ -1264,8 +1422,10 @@ namespace giac {
 	    }
 	  }
 	} // end for heapbeg!=heapend
+#ifdef HEAP_STATS
 	if (debug_infolevel>20)
 	  CERR << CLOCK() << " heap_mult, %age of chains" << count1/total << " " << count2/total << " " << std::endl;
+#endif
 	delete [] heap;
 	return;
       }
@@ -1466,7 +1626,7 @@ namespace giac {
   }
 
 
-  template<class T,class U>
+  template<class T,class U,class R>
   struct threadmult_t {
     const std::vector< T_unsigned<T,U> > * v1ptr ;
     std::vector< typename std::vector< T_unsigned<T,U> >::const_iterator > * v1ptrs, * v2ptrs;
@@ -1474,7 +1634,8 @@ namespace giac {
     U degdiv;
     unsigned current_deg;
     unsigned clock;
-    int reduce,status;
+    R reduce;
+    int status;
     bool use_heap;
     T * prod;
     std::vector< vector_size64< std::pair<unsigned,unsigned> > > * vindexptr;
@@ -1484,11 +1645,11 @@ namespace giac {
 
 #if defined HAVE_PTHREAD_H && defined HAVE_LIBPTHREAD
 
-  template<class T,class U> void * do_threadmult(void * ptr){
-    threadmult_t<T,U> * argptr = (threadmult_t<T,U> *) ptr;
+  template<class T,class U,class R> void * do_threadmult(void * ptr){
+    threadmult_t<T,U,R> * argptr = (threadmult_t<T,U,R> *) ptr;
     argptr->status=1;
     argptr->clock=CLOCK();
-    int reduce=argptr->reduce;
+    R reduce=argptr->reduce;
     const std::vector< T_unsigned<T,U> > * v1 = argptr->v1ptr;
     std::vector< typename std::vector< T_unsigned<T,U> >::const_iterator >  * v2ptrs = argptr->v2ptrs;
     std::vector< T_unsigned<T,U> > & v = *argptr->vptr;
@@ -1641,7 +1802,7 @@ namespace giac {
 	    v.pop_back();
 	  }
 	  else
-	    g=0;
+	    g=T(0);
 	  std::vector< std::pair<unsigned short,unsigned short> >::iterator it,itend;
 	  vector_size32< std::pair<unsigned short,unsigned short> > * vptr;
 	  nouveau.clear();
@@ -1672,11 +1833,12 @@ namespace giac {
 	    // std::pop_heap(heapbeg,heapend);
 	    --heapend;
 	  }
-	  if (g != 0)
+	  if (!is_zero(g))
 	    v.push_back(T_unsigned<T,U>(g,topu));
 	  // erase top node, then push each element of the incremented top chain 
 	  it=nouveau.begin();
 	  itend=nouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    it1=it1beg+it->first;
 	    u1=it1->u;
@@ -1684,10 +1846,16 @@ namespace giac {
 	    d2=v2deg-(d-d1);
 	    it2beg=(*v2ptrs)[d2];
 	    u=u1+(it2beg+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vptr=&(*vsmallindexptr)[previndex];
+	      vptr->push_back(*it);
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=heapend-heapbeg,parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vptr=&(*vsmallindexptr)[heapbeg->v];
+	      vptr=&(*vsmallindexptr)[previndex=heapbeg->v];
 	      vptr->push_back(*it);
 	      continue;
 	    }
@@ -1698,7 +1866,7 @@ namespace giac {
 		break;
 	      if (u==(heapbeg+parentindex)->u){
 		done=true;
-		vptr=&(*vsmallindexptr)[(heapbeg+parentindex)->v];
+		vptr=&(*vsmallindexptr)[previndex=(heapbeg+parentindex)->v];
 		vptr->push_back(*it);
 		break;
 	      }
@@ -1706,7 +1874,7 @@ namespace giac {
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vptr=&(*vsmallindexptr)[heapend->v];
+	      vptr=&(*vsmallindexptr)[previndex=heapend->v];
 	      vptr->clear();
 	      vptr->push_back(*it);
 	      heapend->u=u;
@@ -1730,7 +1898,7 @@ namespace giac {
 	    v.pop_back();
 	  }
 	  else
-	    g=0;
+	    g=T(0);
 	  std::vector< std::pair<unsigned,unsigned> >::iterator it,itend;
 	  vector_size64< std::pair<unsigned,unsigned> > * vptr;
 	  nouveau.clear();
@@ -1761,11 +1929,12 @@ namespace giac {
 	    // std::pop_heap(heapbeg,heapend);
 	    --heapend;
 	  }
-	  if (g != 0)
+	  if (!is_zero(g))
 	    v.push_back(T_unsigned<T,U>(g,topu));
 	  // erase top node, then push each element of the incremented top chain 
 	  it=nouveau.begin();
 	  itend=nouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    it1=it1beg+it->first;
 	    u1=it1->u;
@@ -1773,10 +1942,16 @@ namespace giac {
 	    d2=v2deg-(d-d1);
 	    it2beg=(*v2ptrs)[d2];
 	    u=u1+(it2beg+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vptr=&(*vindexptr)[previndex];
+	      vptr->push_back(*it);
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=heapend-heapbeg,parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vptr=&(*vindexptr)[heapbeg->v];
+	      vptr=&(*vindexptr)[previndex=heapbeg->v];
 	      vptr->push_back(*it);
 	      continue;
 	    }
@@ -1787,7 +1962,7 @@ namespace giac {
 		break;
 	      if (u==(heapbeg+parentindex)->u){
 		done=true;
-		vptr=&(*vindexptr)[(heapbeg+parentindex)->v];
+		vptr=&(*vindexptr)[previndex=(heapbeg+parentindex)->v];
 		vptr->push_back(*it);
 		break;
 	      }
@@ -1795,7 +1970,7 @@ namespace giac {
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vptr=&(*vindexptr)[heapend->v];
+	      vptr=&(*vindexptr)[previndex=heapend->v];
 	      vptr->clear();
 	      vptr->push_back(*it);
 	      heapend->u=u;
@@ -1834,7 +2009,7 @@ namespace giac {
       g1=it1->g;
       it2beg=(*v2ptrs)[d2];
       it2end=(*v2ptrs)[d2+1];
-      if (reduce){
+      if (!is_zero(reduce)){
 	for (it2=it2beg;it2!=it2end;++it2){
 	  u2=it2->u;
 	  u=u1+u2;
@@ -1878,8 +2053,8 @@ namespace giac {
     return &v;
   }
 
-  template<class T,class U>
-  bool threadmult(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,U degdiv,int reduce,size_t possible_size=100){
+  template<class T,class U,class R>
+  bool threadmult(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,U degdiv,const R & reduce,size_t possible_size=100){
     if (!threads_allowed)
       return false;
     v.clear();
@@ -1910,7 +2085,9 @@ namespace giac {
     }
     // if array multiplication is faster, set prod
     bool use_heap = (heap_mult<0) || (heap_mult>0 && v1v2>heap_mult);
-    if (!prod && use_heap && nthreads<2) 
+    if (!prod && use_heap 
+	&& nthreads<2 // ???
+	) // multi-thread heap disabled because of locks by inserting in chains 
       return false;  
     if (debug_infolevel>20){
       CERR << "// " << CLOCK() << "using threaded " ;
@@ -1958,7 +2135,7 @@ namespace giac {
     // will launch deg1v+1 threads to compute each degree
     pthread_t tab[deg1v+1];
     // threadmult_t<T,U> arg[deg1v+1];
-    threadmult_t<T,U> * arg=new threadmult_t<T,U>[deg1v+1];
+    threadmult_t<T,U,R> * arg=new threadmult_t<T,U,R>[deg1v+1];
     possible_size=0;
     int res=0;
     int i=deg1v;
@@ -1995,7 +2172,7 @@ namespace giac {
 	}
 	if (debug_infolevel>30)
 	  CERR << "Computing degree " << i << " " << CLOCK() << std::endl;
-	do_threadmult<T,U>(&arg[i]);
+	do_threadmult<T,U,R>(&arg[i]);
 	threads_time += arg[i].clock;
 	possible_size += arg[i].vptr->size();	
       }
@@ -2028,7 +2205,7 @@ namespace giac {
 	  arg[i].vindexptr=0;
 	  arg[i].vsmallindexptr=0;
 	}
-	res=pthread_create(&tab[i],(pthread_attr_t *) NULL,do_threadmult<T,U>,(void *) &arg[i]);
+	res=pthread_create(&tab[i],(pthread_attr_t *) NULL,do_threadmult<T,U,R>,(void *) &arg[i]);
 	if (res){
 	  // should cancel previous threads and delete created arg[i].vptr
 	  delete [] arg;
@@ -2053,7 +2230,7 @@ namespace giac {
 		arg[i].v1ptrs=&v1it;
 		arg[i].v2ptrs=&v2it;
 		arg[i].vptr = new std::vector< T_unsigned<T,U> >;
-		arg[i].vptr->reserve(arg[k].vptr->size());
+		arg[i].vptr->reserve((3*arg[k].vptr->size())/2); // ???
 		arg[i].degdiv=degdiv;
 		arg[i].current_deg=i;
 		arg[i].reduce=reduce;
@@ -2063,7 +2240,7 @@ namespace giac {
 		arg[i].heapptr=arg[k].heapptr;
 		arg[i].vindexptr=arg[k].vindexptr;
 		arg[i].vsmallindexptr=arg[k].vsmallindexptr;
-		res=pthread_create(&tab[i],(pthread_attr_t *) NULL,do_threadmult<T,U>,(void *) &arg[i]);
+		res=pthread_create(&tab[i],(pthread_attr_t *) NULL,do_threadmult<T,U,R>,(void *) &arg[i]);
 		if (res){
 		  // should cancel previous threads and delete created arg[i].vptr
 		  delete [] arg;
@@ -2150,23 +2327,115 @@ namespace giac {
   }
 
 #else // PTHREAD
-  template<class T,class U>
-  bool threadmult(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,U degdiv,int reduce,size_t possible_size=100){
+  template<class T,class U,class R>
+  bool threadmult(const std::vector< T_unsigned<T,U> > & v1,const std::vector< T_unsigned<T,U> > & v2,std::vector< T_unsigned<T,U> > & v,U degdiv,const R& reduce,size_t possible_size=100){
     return false;
   }
 
 #endif // PTHREAD
 
+  template<class U>
+  void partial_degrees2vars(const index_t & d,std::vector<U> & vars){
+    int dim=int(d.size());
+    vars[dim-1]=1;
+    for (int i=dim-2;i>=0;--i){
+      vars[i]=(1+d[i+1])*vars[i+1];
+    }
+  }
+
+  template<class U>
+  void partial_degrees(U u,const std::vector<U> & vars,index_t & res){
+    for (int k=int(vars.size())-1;k>0;--k){
+      U v=u%vars[k-1];
+      res[k]=v/vars[k];
+    }
+    res[0]=u/vars[0];
+  }
+
+  template<class T,class U>
+  void partial_degrees(const std::vector< T_unsigned<T,U> > & a,const std::vector<U> & vars,index_t & res){
+    typename std::vector< T_unsigned<T,U> >::const_iterator it=a.begin(),itend=a.end();
+    int dim=int(vars.size());
+    res.clear(); res.resize(dim);
+    index_t cur(dim);
+    for (;it!=itend;++it){
+      partial_degrees(it->u,vars,cur);
+      index_lcm(res,cur,res);
+    }
+  }
+
+  // convert u monomial from vars source to vars target, tmp is a temp index_t
+  template<class U> 
+  void convert(U & u,const std::vector<U> & source,const std::vector<U> & target,index_t & tmp){
+    partial_degrees(u,source,tmp);
+    u=0;
+    for (int i=int(source.size())-1;i>=0;--i){
+      u += target[i]*tmp[i];
+    }
+  }
+
+  template<class T,class U>
+  void convert(std::vector< T_unsigned<T,U> > & a,const std::vector<U> & source,const std::vector<U> & target){
+    index_t tmp(source.size());
+    typename std::vector< T_unsigned<T,U> >::iterator it=a.begin(),itend=a.end();
+    for (;it!=itend;++it){
+      convert(it->u,source,target,tmp);
+    }
+  }
+
+  inline bool hashdivrem_finish_later(longlong a){ return false;}
+  inline bool hashdivrem_finish_later(double a){return false;}
+  inline bool hashdivrem_finish_later(int a){ return false; }
+  inline bool hashdivrem_finish_later(const vecteur & v){ return false; }
+  inline bool hashdivrem_finish_later(const std::vector<int> & v){ return false; }
+#ifdef INT128
+  inline bool hashdivrem_finish_later(int128_t a){return false;}
+#endif
+  
+  inline bool hashdivrem_finish_later(const gen & a){return true;}
+  inline bool hashdivrem_finish_later(const my_mpz & a){return true;}
+#ifdef HAVE_GMPXX_H
+  inline bool hashdivrem_finish_later(const mpz_class & a){return true;}
+#endif
+
+  template<class U>
+  inline bool one_index_smaller(U u,U v,const std::vector<int> & varsshift){
+    if (u<v) 
+      return true;
+    return false;
+    /*
+    // the code below is too slow
+    std::vector<int>::const_iterator it=varsshift.begin(),itend=varsshift.end();
+    for (;it!=itend;++it){
+      int shift=*it;
+      U u1=(u>>shift);
+      U v1=(v>>shift);
+      if (u1<v1) 
+	return true;
+      u -= (u1 << shift);
+      v -= (v1 << shift);
+    }
+    return false;
+    */
+  }
+
+  // #define HEAP_STATS
   // note that U may be of type vector of int or an int
   // + is used to multiply monomials and - to divide
   // / should return the quotient of the main variable exponent
   // > should return true if a monomial has main degree >
   // vars is the list of monomials x,y,z,etc. as translated in U type
-  // quo_only==2 means we know that b divides a and we search the cofactor
-  // quo_only==1 means we want to check that b divides a
-  // quo_only==-1 means compute quotient first using heap div then r=a-b*q
+  // quo_only==-3 means heap div (compute quo and rem)
   // quo_only==-2 means compute quotient only using heap div
-  // if quo_only is >= 2 or <=-2, r is not computed
+  // quo_only==-1 heap quotient then guess between heap remainder 
+  //              or r=a-b*q must be done by caller (if returns 2)
+  // quo_only==0 array division or univariate with hashmap
+  // quo_only==1 means we want to check that b divides a
+  // quo_only==2 means we know that b divides a and we search the cofactor
+  // quo_only==3 array division if enough memory
+  // quo_only>3 univariate division with hashmap or map
+  // for coefficients monomial storage
+  // returns 1 if ok, 2 if ok but remainder not computed, 0 or -1 otherwise
   template<class T,class U,class R>
   int hashdivrem(const std::vector< T_unsigned<T,U> > & a,const std::vector< T_unsigned<T,U> > & b,std::vector< T_unsigned<T,U> > & q,std::vector< T_unsigned<T,U> > & r,const std::vector<U> & vars,const R & reduce,double qmax,bool allowrational,int quo_only=0){
     // CERR << "hashdivrem dim " << vars.size() << " clock " << CLOCK() << std::endl;
@@ -2192,7 +2461,7 @@ namespace giac {
     }
     U rstop=U(bdeg) << mainvar;
     typename std::vector< T_unsigned<T,U> >::iterator it1,it1end;
-    typename std::vector< T_unsigned<T,U> >::const_iterator it2,it2end,itbbeg=b.begin(),itbend=b.end(),ita=a.begin(),itaend=a.end(),cit,citend;
+    typename std::vector< T_unsigned<T,U> >::const_iterator it2,it2end,itbbeg=b.begin(),itbend=b.end(),ita=a.begin(),itaend=a.end(),cit,citend,qbeg;
     T binv=b.front().g;
     if (!is_zero(reduce))
       binv=invmod(binv,reduce);
@@ -2222,12 +2491,14 @@ namespace giac {
     unsigned as=unsigned(a.size()),bs=unsigned(b.size());
     double v1v2=double(as)*bs;
     // FIXME, if bdeg==0
-    if ( (!quo_only || quo_only==3) && 
+    if (
+	(!quo_only || quo_only==3) &&
+	//quo_only==3 && //is_zero(reduce) &&
 	bdeg && as>=a.front().u/25. // array div disabled, probably too much memory used
 	&& heap_mult>=0 && a.front().u < 512e6/sizeof(T)){
       U umax=a.front().u,u;
-      if (debug_infolevel>20)
-	CERR << "array division, a size " << a.size() << " b size " << b.size() << " u " << umax << std::endl;
+      if (debug_infolevel>1)
+	CERR << CLOCK()*1e-6 << " array division, a size " << a.size() << " b size " << b.size() << " u " << umax << std::endl;
       // array division
       T * rem = new T[unsigned(umax+1)];
       for (u=0;u<=umax;++u)
@@ -2283,7 +2554,7 @@ namespace giac {
 	}
 	else {
 	  int recdivres=hashdivrem(maincoeff,lcoeffb,quo,tmp,std::vector<U>(vars.begin()+1,vars.end()),reduce,qmax,allowrational);
-	  if (recdivres!=1){
+	  if (recdivres<1){
 	    delete [] rem;
 	    return recdivres;
 	  }
@@ -2332,15 +2603,31 @@ namespace giac {
       }
       delete [] rem;
       return 1;
-    }
+    } // end array division
     bool use_heap=false && 
       (heap_mult>0 
        && v1v2>=heap_mult
        );
 #if 1 // heap division
-    if (heap_mult<0 || use_heap || (quo_only && quo_only!=3)){
-      if (debug_infolevel>20)
-	CERR << "heap division, a size " << a.size() << " b size " << b.size() << " vars " << vars << std::endl;
+    if (heap_mult<0 || use_heap ||
+	(quo_only && quo_only<3)){
+	  // quo_only<3){
+      bool norem=quo_only==2 || quo_only==-2;
+      norem=false; // currently disabled, it's not clear it wins something
+      std::vector<int> varsshift; // vars=2^varsshift
+      if (norem){
+	for (size_t i=0;i<vars.size();++i){
+	  int j=sizeinbase2(vars[i])-1;
+	  if (vars[i]!=(U(1)<<j))
+	    norem=false;
+	  varsshift.push_back(j);
+	}
+      }
+#ifdef HEAP_STATS
+      unsigned chain=0,nochain=0,typeopreduce=0,nullq=0;
+#endif
+      if (debug_infolevel>1)
+	CERR << CLOCK()*1e-6 << " heap division, a size " << a.size() << " b size " << b.size() << " vars " << vars << std::endl;
       // heap division:
       // ita an iterator on a, initial value a.begin()
       // a heap with the current state of q*b, initialized to empty heap
@@ -2355,7 +2642,7 @@ namespace giac {
       //
       // vindex[i] is the heap chain corresponding to index i
       // it contains pairs of index in g and q
-      std::vector< std::vector< std::pair<unsigned,unsigned> > > vindex(bs);
+      std::vector< std::vector< std::pair<unsigned,unsigned> > > vindex(bs) ; // ,std::vector< std::pair<unsigned,unsigned> >(4));
       U_unsigned<U> * heap = new U_unsigned<U>[bs], * heapbeg =heap, * heapend=heap ;
       T g;
       U heapu,u;
@@ -2373,7 +2660,7 @@ namespace giac {
 	if (heapbeg!=heapend){
 	  heapu=heapbeg->u;
 	  if (ita!=itaend && ita->u>heapu){
-	    if (ita->u>bu){
+	    if (ita->u>=bu){
 	      heapu=ita->u;
 	      g=ita->g;
 	      ++ita;
@@ -2385,16 +2672,24 @@ namespace giac {
 	      std::vector< std::pair<unsigned,unsigned> >::iterator it,itend;
 	      nouveau.clear();
 	      while (heapend!=heapbeg && heapu==heapbeg->u){
-		nouveau.clear();
+		//nouveau.clear();
 		// add all elements of the top chain	
-		it=vindex[heapbeg->v].begin();
-		itend=vindex[heapbeg->v].end();
+		std::vector< std::pair<unsigned,unsigned> > & V=vindex[heapbeg->v];
+		it=V.begin();
+		itend=V.end();
+		qbeg=q.begin();
+		size_t qsize=q.size();
 		for (;it!=itend;++it){
-		  type_operator_plus_times_reduce((itbbeg+it->first)->g,(q.begin()+it->second)->g,g,reduce);
+#ifdef HEAP_STATS
+		  ++typeopreduce;
+#endif
+		  unsigned & its=it->second;
+		  type_operator_plus_times_reduce((itbbeg+it->first)->g,(qbeg+its)->g,g,reduce);
 		  // increment 2nd poly index of the elements of the top chain
-		  ++it->second;
-		  if (it->second<q.size())
+		  ++its;
+		  if (its<qsize){
 		    nouveau.push_back(*it);
+		  }
 		  else // wait for computation of a new term of a before adding to the heap 
 		    qnouveau.push_back(*it);
 		}
@@ -2406,33 +2701,47 @@ namespace giac {
 #endif
 		// std::pop_heap(heapbeg,heapend);
 		--heapend;
+	      } // while heapend!=heapbeg && heapu==
+	      {
 		// push each element of the incremented top chain 
 		it=nouveau.begin();
 		itend=nouveau.end();
-		for (;it!=itend;++it){
-		  u=(itbbeg+it->first)->u+(q.begin()+it->second)->u;
+		for (;it!=itend;++it) {
+		  u=(itbbeg+it->first)->u+(qbeg+it->second)->u;
+		  if (norem && one_index_smaller(u,bu,varsshift)) continue;
 		  // check if u is in the path to the root of the heap
 		  unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 		  if (holeindex && u==heapbeg->u){
+#ifdef HEAP_STATS
+		    ++chain;
+#endif
 		    vindex[heapbeg->v].push_back(*it);
 		    continue;
 		  }
 		  bool done=false;
 		  while (holeindex){
 		    parentindex=(holeindex-1) >> 1;
-		    if (u<(heapbeg+parentindex)->u)
+		    U pu=(heapbeg+parentindex)->u;
+		    if (u<pu)
 		      break;
-		    if (u==(heapbeg+parentindex)->u){
+		    if (u==pu){
 		      done=true;
 		      vindex[(heapbeg+parentindex)->v].push_back(*it);
+#ifdef HEAP_STATS
+		      ++chain;
+#endif
 		      break;
 		    }
 		    holeindex=parentindex;
 		  }
 		  if (!done){
+#ifdef HEAP_STATS
+		    ++nochain;
+#endif
 		    // not found, add a new node to the heap
-		    vindex[heapend->v].clear();
-		    vindex[heapend->v].push_back(*it);
+		    std::vector< std::pair<unsigned,unsigned> > & V=vindex[heapend->v];
+		    V.clear();
+		    V.push_back(*it);
 		    heapend->u=u;
 		    ++heapend;
 #ifdef USTL
@@ -2445,18 +2754,24 @@ namespace giac {
 		} // end adding incremented pairs from nouveau
 	      } // end loop on monomial of the heap having the same u
 	      if (heapu==ita->u){
-		g = ita->g -g ;
+		g = ita->g -g ; // FIXME must be reduced!
+		if (!is_zero(reduce))
+		  g = g % reduce;
 		++ita;
 	      }
 	      else
 		g=-g;
-	      if (is_zero(g))
+	      if (is_zero(g)){
+#ifdef HEAP_STATS
+		++nullq;
+#endif
 		continue;
+	      }
 	    } // end if (heapu>=bu)
 	  } // end else ita->u>heapu
 	} // if heap non empty
 	else {
-	  if (ita!=itaend && (heapu=ita->u)>bu){
+	  if (ita!=itaend && (heapu=ita->u)>=bu){
 	    g=ita->g;
 	    ++ita;
 	  } 
@@ -2472,7 +2787,7 @@ namespace giac {
 	  type_operator_reduce(g,binv,g,reduce);
 	else
 	  g=g/binv;
-	if (qmax && g>qmax){
+	if (qmax && (g>qmax || -g>qmax)){
 	  delete [] heap;
 	  return -1;
 	}
@@ -2483,32 +2798,53 @@ namespace giac {
 	std::vector< std::pair<unsigned,unsigned> >::iterator it,itend;
 	it=qnouveau.begin();
 	itend=qnouveau.end();
+	U prevu=0; int previndex=-1;
 	for (;it!=itend;++it){
 	  if (!it->first) // leading term of b already taken in account
 	    continue;
 	  u=(itbbeg+it->first)->u+(q.begin()+it->second)->u;
+	  if (norem && one_index_smaller(u,bu,varsshift)) continue;
+	  if (u==prevu && previndex>=0){
+#ifdef HEAP_STATS
+	    ++chain;
+#endif
+	    vindex[previndex].push_back(*it);
+	    continue;
+	  }
+	  prevu=u;
 	  // check if u is in the path to the root of the heap
 	  unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 	  if (holeindex && u==heapbeg->u){
-	    vindex[heapbeg->v].push_back(*it);
+#ifdef HEAP_STATS
+	    ++chain;
+#endif
+	    vindex[previndex=heapbeg->v].push_back(*it);
 	    continue;
 	  }
 	  bool done=false;
 	  while (holeindex){
 	    parentindex=(holeindex-1) >> 1;
-	    if (u<(heapbeg+parentindex)->u)
+	    U pu=(heapbeg+parentindex)->u;
+	    if (u<pu)
 	      break;
-	    if (u==(heapbeg+parentindex)->u){
+	    if (u==pu){
+#ifdef HEAP_STATS
+	      ++chain;
+#endif
 	      done=true;
-	      vindex[(heapbeg+parentindex)->v].push_back(*it);
+	      vindex[previndex=(heapbeg+parentindex)->v].push_back(*it);
 	      break;
 	    }
 	    holeindex=parentindex;
 	  }
 	  if (!done){
+#ifdef HEAP_STATS
+	    ++nochain;
+#endif
 	    // not found, add a new node to the heap
-	    vindex[heapend->v].clear();
-	    vindex[heapend->v].push_back(*it);
+	    std::vector< std::pair<unsigned,unsigned> > & V=vindex[(previndex=heapend->v)];
+	    V.clear();
+	    V.push_back(*it);
 	    heapend->u=u;
 	    ++heapend;
 #ifdef USTL
@@ -2521,30 +2857,56 @@ namespace giac {
 	} // end adding incremented pairs from qnouveau
 	qnouveau.clear();
       } // for (;;)
-      // r still empty
-      if (debug_infolevel>20)
-	CERR << "Finished computing quotient, size " << q.size() << " " << CLOCK() << std::endl ;
-      if (quo_only>=2 || quo_only<=-2){
-	delete [] heap;
-	return 1;
-      }
-      // do not define QUO_ONLY unless R==int
-#ifdef QUO_ONLY
-      if (quo_only==-1 && is_zero(reduce) 
-	  && q.size()*double(b.size())>10000 
-	  ){
-	if (debug_infolevel>20)
-	  CERR << "Computing q*b, size " << q.size() << " " << b.size() << std::endl ;
-	if (debug_infolevel>30)
-	  CERR << "Computing b*q, " << b << " * " << q << " " << vars.front() << " " <<a.size() << std::endl ;
-	std::vector< T_unsigned<T,U> > bq;
-	if (!threadmult(b,q,bq,vars.front(),0,a.size()))
-	  smallmult(b,q,bq,reduce,as);
-	smallsub(a,bq,r);
-	delete [] heap;
-	return 1;
-      }
+#ifdef HEAP_STATS
+      if (debug_infolevel)
+	CERR << "chain " << chain << ", nochain " << nochain << ", type_op_reduce " << typeopreduce << " null quotients" << nullq << std::endl;
 #endif
+      // r still empty
+      if (debug_infolevel>2)
+	CERR << CLOCK()*1e-6 << " Finished computing quotient, size " << q.size() << std::endl ;
+      if (quo_only==2 || quo_only==-2){
+	delete [] heap;
+	return 1;
+      }
+      if (quo_only==-1 ){
+	// try to compare heap mult and array mult, return for array mult only
+	double qb=double(q.size())*b.size();
+	qb /= a.size();
+	if (debug_infolevel>1)
+	  CERR << CLOCK()*1e-6 << " qb=" << qb << std::endl;
+	if (qb>100){
+	  // the coefficients might be not optimal (mpz_class instead of int)
+	  if (hashdivrem_finish_later(a.front().g)){
+	    delete [] heap;
+	    return 2;
+	  }
+	  // the monomials are not stored efficiently for array *, compress
+	  index_t adeg,bdeg,qdeg,bqdeg;
+	  partial_degrees(a,vars,adeg);
+	  partial_degrees(b,vars,bdeg);
+	  partial_degrees(q,vars,qdeg);
+	  bqdeg=bdeg+qdeg;
+	  bqdeg=index_lcm(bqdeg,adeg);
+	  std::vector<U> newvars(vars.size());
+	  partial_degrees2vars(bqdeg,newvars);
+	  std::vector< T_unsigned<T,U> > acopy(a),bcopy(b),qcopy(q),bq;
+	  convert(acopy,vars,newvars);
+	  convert(bcopy,vars,newvars);
+	  convert(qcopy,vars,newvars);
+	  if (debug_infolevel>1)
+	    CERR << CLOCK()*1e-6 << " compress monomials done" <<std::endl;
+	  if (!threadmult(bcopy,qcopy,bq,newvars.front(),reduce,a.size()))
+	    smallmult(bcopy,qcopy,bq,reduce,as);
+	  smallsub(acopy,bq,r);
+	  if (debug_infolevel>1)
+	    CERR << CLOCK()*1e-6 << " uncompress monomials" <<std::endl;
+	  convert(r,newvars,vars);
+	  if (debug_infolevel>1)
+	    CERR << CLOCK()*1e-6 << " uncompress monomials end"<< std::endl;
+	  delete [] heap;
+	  return 1;
+	}
+      }
       // now q is computed, combine a and remaining product to r
       for (;heapbeg!=heapend;){
 	heapu=heapbeg->u;
@@ -2568,10 +2930,11 @@ namespace giac {
 	  it=vindex[heapbeg->v].begin();
 	  itend=vindex[heapbeg->v].end();
 	  for (;it!=itend;++it){
-	    type_operator_plus_times_reduce((itbbeg+it->first)->g,(q.begin()+it->second)->g,g,reduce);
+	    unsigned & its=it->second;
+	    type_operator_plus_times_reduce((itbbeg+it->first)->g,(q.begin()+its)->g,g,reduce);
 	    // increment 2nd poly index of the elements of the top chain
-	    ++it->second;
-	    if (it->second<q.size())
+	    ++its;
+	    if (its<q.size())
 	      qnouveau.push_back(*it);
 	  }
 	  // erase top node, 
@@ -2585,30 +2948,38 @@ namespace giac {
 	  // push each element of the incremented top chain 
 	  it=qnouveau.begin();
 	  itend=qnouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    u=(itbbeg+it->first)->u+(q.begin()+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vindex[previndex].push_back(*it);
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vindex[heapbeg->v].push_back(*it);
+	      vindex[(previndex=heapbeg->v)].push_back(*it);
 	      continue;
 	    }
 	    bool done=false;
 	    while (holeindex){
 	      parentindex=(holeindex-1) >> 1;
-	      if (u<(heapbeg+parentindex)->u)
+	      U pu=(heapbeg+parentindex)->u;
+	      if (u<pu)
 		break;
-	      if (u==(heapbeg+parentindex)->u){
+	      if (u==pu){
 		done=true;
-		vindex[(heapbeg+parentindex)->v].push_back(*it);
+		vindex[(previndex=(heapbeg+parentindex)->v)].push_back(*it);
 		break;
 	      }
 	      holeindex=parentindex;
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vindex[heapend->v].clear();
-	      vindex[heapend->v].push_back(*it);
+	      std::vector< std::pair<unsigned,unsigned> > & V=vindex[(previndex=heapend->v)];
+	      V.clear();
+	      V.push_back(*it);
 	      heapend->u=u;
 	      ++heapend;
 #ifdef USTL
@@ -2692,7 +3063,7 @@ namespace giac {
       }
       else {
 	int recdivres=hashdivrem(maincoeff,lcoeffb,quo,tmp,std::vector<U>(vars.begin()+1,vars.end()),reduce,qmax,allowrational);
-	if (recdivres!=1)
+	if (recdivres<1)
 	  return recdivres;
 	if (!tmp.empty())
 	  return 0;
